import 'dart:async';

import 'package:analyzer/dart/element/element.dart';
import 'package:b_annotations/b_annotations.dart';
import 'package:b_generator/src/helpers/string_helpers.dart';
import 'package:build/build.dart';
import 'package:source_gen/source_gen.dart';

import 'helpers/annotated_method_visitor.dart';

class BlocGenerator extends GeneratorForAnnotation<BlocAnnotation> {
  @override
  FutureOr<String> generateForAnnotatedElement(Element element, ConstantReader annotation, BuildStep buildStep) {
    final Set<MethodElement> events = {};
    element.visitChildren(AnnotatedMethodVisitor<EventAnnotation>((method) => events.add(method)));

    return """
    abstract class GeneratedBloc extends Bloc<${eventName(element)}, ${stateName(element)}> {
      ${stateName(element)} get initialState;
      
      ${mapEventsToState(element, events)}
      
      ${events.map((event) => mapEventToState(element, event)).join('\n')}
    }""";
  }

  String eventName(Element element) => '${blocName(element)}Event';

  String stateName(Element element) => '${blocName(element)}State';

  String blocName(Element element) => element.name;

  String mapEventsToState(Element element, Set<MethodElement> events) {
    return '''
    @override
    Stream<${stateName(element)}> mapEventToState(final ${eventName(element)} event) async* {
        ${events.map((event) => 'if (event is ${className(event.name)}) yield* ${event.name}(${event.parameters.map((parameter) => 'event.${parameter.name}').join(', ')});').join('\n')}
    }
    ''';
  }

  mapEventToState(Element element, MethodElement event) {
    return 'Stream<${stateName(element)}> ${event.name}(${event.parameters.map((parameter) => 'final ${parameter.type} ${parameter.name}').join(', ')});';
  }
}
